#!/bin/python
from __future__ import absolute_import, division, print_function
import numpy as np
from gwmemory import harmonics
import glob2 as glob
import pkg_resources
import pandas as pd


def gamma(lm1, lm2, incs=None, theta=None, phi=None, y_lmlm_factor=None):
    """
    Coefficients mapping the spherical harmonic components of the oscillatory strain to the memory.

    Computed according to equation XYZ of Talbot et al. (2018), arXiv:18XX:ABCDE.
    Output modes with l=range(2, 20), m=m1-m2.

    Parameters
    ----------
    lm1: str
        first input spherical harmonic mode
    lm2: str
        second input spherical haromonic mode
    incs: array, optional
        observer inclination values over which to compute the final integral
    theta: array, optional
        1d array of binary inclination values, over which to compute first integral
    phi: array, optional
        1d array of binary polarisation values, over which to compute the first integral
    y_lmlm_factor: array, optional
        Array over of spherical harmonic factor evaluated on meshgrid of theta, phi

    Return
    ------
    list: Coefficients for output modes, l=range(2, 20), m=m1-m2
    """
    l1, m1 = int(lm1[0]), int(lm1[1:])
    l2, m2 = int(lm2[0]), int(lm2[1:])

    if incs is None:
        incs = np.linspace(0, np.pi, 500)
    pol = 0
    if theta is None:
        theta = np.linspace(0, np.pi, 250)
    if phi is None:
        phi = np.linspace(0, 2 * np.pi, 500)

    if y_lmlm_factor is None:
        s = -2
        th, ph = np.meshgrid(theta, phi)

        l1, m1 = int(lm1[0]), int(lm1[1:])
        l2, m2 = int(lm2[0]), int(lm2[1:])

        y_lmlm_factor = harmonics.sYlm(s, l1, m1, th, ph) * (-1)**(l2+m2) * harmonics.sYlm(-s, l2, -m2, th, ph)

    lambda_lm1_lm2 = np.array([lambda_lmlm(inc, pol, lm1, lm2, theta, phi, y_lmlm_factor) for inc in incs])

    sin_inc = np.sin(-incs)

    harm = {}
    for l, m in harmonics.lmax_modes(20):
        harm['{}{}'.format(l, m)] = harmonics.sYlm(-2, l, m, incs, pol)

    ells = np.arange(2, 21, 1)

    delta_m = m1 - m2
    result = []
    for ell in ells:
        if ell < abs(delta_m):
            result.append(0)
        else:
            result.append(np.trapz(lambda_lm1_lm2 *
                                   np.conjugate(harm['{}{}'.format(ell, delta_m)])
                                   * sin_inc, incs) * 2 * np.pi)

    return result


def lambda_matrix(inc, pol, lm1, lm2, theta=None, phi=None, y_lmlm_factor=None):
    """
    Angular integral for a specific ll'mm' as given by equation XYZ of Talbot et al. (2018), arXiv:18XX:ABCDE.

    The transverse traceless part of the integral over all binary orientations is returned.

    The integral is given by:
    \int_{S^{2}} d\Omega' Y^{-2}_{\ell_1 m_1}(\Omega') \bar{Y}^{-2}_{\ell_2 m_2}(\Omega') \times \\
    \left[\frac{n_jn_k}{1-n_{l}N_{l}} \right]^{TT}

    Parameters
    ----------
    inc: float
        observer inclination
    pol: float
        observer polarisation
    lm1: str
        first lm value format is e.g., '22'
    lm2: str
        second lm value format is e.g., '22'
    theta: array, optional
        1d array of binary inclination values, over which to integrate
    phi: array, optional
        1d array of binary polarisation values, over which to integrate
    y_lmlm_factor: array, optional
        Array over of spherical harmonic factor evaluated on meshgrid of theta, phi

    Return
    ------
    lambda_mat: array
        three by three transverse traceless matrix of the appropriate integral
    """
    if theta is None:
        theta = np.linspace(0, np.pi, 250)
    if phi is None:
        phi = np.linspace(0, 2 * np.pi, 500)

    if y_lmlm_factor is None:
        ss = -2

        th, ph = np.meshgrid(theta, phi)

        l1, m1 = int(lm1[0]), int(lm1[1:])
        l2, m2 = int(lm2[0]), int(lm2[1:])

        y_lmlm_factor = harmonics.sYlm(ss, l1, m1, th, ph) * (-1)**(l2+m2) * harmonics.sYlm(-ss, l2, -m2, th, ph)

    n = [np.outer(np.cos(phi), np.sin(theta)), np.outer(np.sin(phi), np.sin(theta)),
         np.outer(np.ones_like(phi), np.cos(theta))]
    N = [np.sin(inc) * np.cos(pol), np.sin(inc) * np.sin(pol), np.cos(inc)]
    n_dot_N = sum(n_i * N_i for n_i, N_i in zip(n, N))
    n_dot_N[n_dot_N == 1] = 0
    denominator = 1/(1-n_dot_N)

    sin_array = np.outer(phi**0, np.sin(theta))

    angle_integrals_r = np.zeros((3, 3))
    angle_integrals_i = np.zeros((3, 3))
    for j in range(3):
        for k in range(j + 1):
            # projection done here to avoid divergences
            integrand = (n[j] * n[k] - (n[j] * N[k] + n[k] * N[j]) * n_dot_N + N[j] * N[k] * n_dot_N**2)\
                        * sin_array * denominator * y_lmlm_factor
            angle_integrals_r[j, k] = np.trapz(np.trapz(np.real(integrand), theta), phi)
            angle_integrals_i[j, k] = np.trapz(np.trapz(np.imag(integrand), theta), phi)
            angle_integrals_r[k, j] = np.trapz(np.trapz(np.real(integrand), theta), phi)
            angle_integrals_i[k, j] = np.trapz(np.trapz(np.imag(integrand), theta), phi)

    proj = np.identity(3) - np.outer(N, N)
    lambda_mat = angle_integrals_r + 1j * angle_integrals_i
    lambda_mat -= proj*np.trace(lambda_mat) / 2

    return lambda_mat


def lambda_lmlm(inc, pol, lm1, lm2, theta=None, phi=None, y_lmlm_factor=None):
    """
    Angular integral for a specific ll'mm' as given by equation XYZ of Talbot et al. (2018), arXiv:18XX:ABCDE.

    The transverse traceless part of the integral over all binary orientations is returned.

    The integral is given by:
    \frac{1}{2} \int_{S^{2}} d\Omega' Y^{-2}_{\ell_1 m_1}(\Omega') \bar{Y}^{-2}_{\ell_2 m_2}(\Omega') \times \\
    \left[\frac{n_jn_k}{1-n_{l}N_{l}} \right]^{TT} (e^{+}_{jk} - i e^{\times}_{jk})

    Parameters
    ----------
    inc: float
        observer inclination
    pol: float
        observer polarisation
    lm1: str
        first lm value format is e.g., '22'
    lm2: str
        second lm value format is e.g., '22'
    theta: array, optional
        1d array of binary inclination values, over which to integrate
    phi: array, optional
        1d array of binary polarisation values, over which to integrate
    y_lmlm_factor: array, optional
        Array over of spherical harmonic factor evaluated on meshgrid of theta, phi

    Return
    ------
    float, complex: lambda_lmlm (lambda_plus - i lambda_cross)

    """
    lambda_mat = lambda_matrix(inc, pol, lm1, lm2, theta, phi, y_lmlm_factor)

    plus, cross = omega_ij_to_omega_pol(lambda_mat, inc, pol)

    return (plus - 1j * cross) / 2


def omega_ij_to_omega_pol(omega_ij, inc, pol):
    """
    Map from strain tensor to plus and cross modes.

    We assume that only plus and cross are present.

    Parameters
    ----------
    omega_ij: array
        3x3 matrix describing strain or a proxy for strain
    inc: float
        inclination of source
    pol: float
        polarisation of source

    Returns
    -------
    hp, hx - (complex) time series
    """
    theta, phi, psi = inc, pol, 0.

    wx, wy, wz = wave_frame(theta, phi, psi)

    omega_plus = np.einsum('ij,ij->', omega_ij, plus_tensor(wx, wy, wz))
    omega_cross = np.einsum('ij,ij->', omega_ij, cross_tensor(wx, wy, wz))

    return omega_plus, omega_cross


def plus_tensor(wx, wy, wz=np.array([0, 0, 1])):
    """Calculate the plus polarization tensor for some basis.c.f., eq. 2 of https://arxiv.org/pdf/1710.03794.pdf"""
    e_plus = np.outer(wx, wx) - np.outer(wy, wy)
    return e_plus


def cross_tensor(wx, wy, wz=np.array([0, 0, 1])):
    """Calculate the cross polarization tensor for some basis.c.f., eq. 2 of https://arxiv.org/pdf/1710.03794.pdf"""
    e_cross = np.outer(wx, wy) + np.outer(wy, wx)
    return e_cross


def wave_frame(theta, phi, psi=0.):
    """generate wave-frame basis from three angles, see Nishizawa et al. (2009)"""
    cth, sth = np.cos(theta), np.sin(theta)
    cph, sph = np.cos(phi), np.sin(phi)
    cps, sps = np.cos(psi), np.sin(psi)

    u = np.array([cph * cth, cth * sph, -sth])
    v = np.array([-sph, cph, 0])

    wx = -u * sps - v * cps
    wy = -u * cps + v * sps
    wz = np.cross(wx, wy)

    return wx, wy, wz


def load_gamma(data_dir=None):
    if data_dir is None:
        data_dir = pkg_resources.resource_filename(__name__, 'data')
    data_files = glob.glob('{}/gamma*.dat'.format(data_dir))
    gamma_lmlm = {}
    for file_name in data_files:
        delta_m = file_name.split('_')[-1][:-4]
        gamma_lmlm[delta_m] = pd.read_csv(file_name, sep='\t')
    return gamma_lmlm



import numpy as np
import deepdish


# Constants for conversions
cc = 299792458.0  # speed of light in m/s
GG = 6.67384e-11  # Newton in m^3 / (kg s^2)
solar_mass = 1.98855 * 10 ** 30  # solar mass in  kg
kg = 1 / solar_mass
metre = cc**2 / (GG * solar_mass)
secs = cc * metre

Mpc = 3.08568e+22  # Mpc in metres


def m12_to_mc(m1, m2):
    """convert m1 and m2 to chirp mass"""
    return (m1*m2)**(3./5.) / (m1 + m2)**(1./5.)


def m12_to_symratio(m1, m2):
    # convert m1 and m2 to symmetric mass ratio
    return m1 * m2 / (m1 + m2)**2


def mc_eta_to_m12(mc, eta):
    """
    Convert chirp mass and symmetric mass ratio to component masses.

    Input: mc - chirp mass
    eta - symmetric mass ratio
    Return: m1, m2 - primary and secondary masses, m1>m2
    """
    m1 = mc/eta**0.6*(1+(1-4*eta)**0.5)/2
    m2 = mc/eta**0.6*(1-(1-4*eta)**0.5)/2
    return m1, m2


def m_sol_to_geo(mm):
    """convert from solar masses to geometric units"""
    return mm / kg * GG / cc**2


def m_geo_to_sol(mm):
    """convert from geometric units to solar masses"""
    return mm * kg / GG * cc**2


def time_s_to_geo(time):
    """convert time from seconds to geometric units"""
    return time * cc


def time_geo_to_s(time):
    """convert time from seconds to geometric units"""
    return time / cc


def freq_Hz_to_geo(freq):
    """convert freq from Hz to geometric units"""
    return freq / cc


def freq_geo_to_Hz(freq):
    """convert freq from geometric units to Hz"""
    return freq * cc


def dist_Mpc_to_geo(dist):
    """convert distance from Mpc to geometric units (i.e., metres)"""
    return dist * Mpc


def nfft(ht, sampling_frequency):
    """
    performs an FFT while keeping track of the frequency bins
    assumes input time series is real (positive frequencies only)

    ht = time series
    sampling_frequency = sampling frequency

    returns
    hf = single-sided FFT of ft normalised to units of strain / sqrt(Hz)
    f = frequencies associated with hf
    """
    # add one zero padding if time series does not have even number of sampling times
    if np.mod(len(ht), 2) == 1:
        ht = np.append(ht, 0)
    LL = len(ht)
    # frequency range
    ff = sampling_frequency / 2 * np.linspace(0, 1, int(LL/2+1))

    # calculate FFT
    # rfft computes the fft for real inputs
    hf = np.fft.rfft(ht)

    # normalise to units of strain / sqrt(Hz)
    hf = hf / sampling_frequency

    return hf, ff


def load_sxs_waveform(file_name, modes=None, extraction='OutermostExtraction.dir'):
    """
    Load the spherical harmonic modes of an SXS numerical relativity waveform.

    Parameters
    ----------
    file_name: str
        Name of file to be loaded.
    modes: dict
        Dictionary of spherical harmonic modes to extract, default is all in ell<=4.
    extraction: str
        String representing extraction method, default is 'OutermostExtraction.dir'
    Returns
    -------
    output: dict
        Dictionary of requested spherical harmonic modes.
    """
    waveform = deepdish.io.load(file_name)
    output = dict()
    mode_array = np.array(())
    if modes is None:
        for ell in range(2, 5):
            for mm in range(-ell, ell + 1):
                mode_array = waveform[extraction]['Y_l{}_m{}.dat'.format(ell, mm)[:, 1:]]
                output[(ell, mm)] = mode_array[:, 1] + 1j * mode_array[:, 2]
    else:
        for mode in modes:
            mode_array = waveform[extraction]['Y_l{}_m{}.dat'.format(mode[0], mode[1])]
            output[mode] = mode_array[:, 1] + 1j * mode_array[:, 2]
    times = mode_array[:, 0]
    return output, times



from . import angles
from . import waveforms
from . import harmonics
from . import qnms
from . import utils
from . import gwmemory
from .gwmemory import time_domain_memory, frequency_domain_memory

name = "gwmemoritz"



# Paul Lasky
# This code package includes code for calculating the properties of quasinormal
# black hole modes
from __future__ import print_function, division, absolute_import

import numpy as np
from gwmemory.harmonics import sYlm
from gwmemory import utils
import os
import pkg_resources


def freq_damping_obs(MM, jj, ell=2, mm=2, nn=0):
    """
    Calculate the quasinormal mode freq and damping time for a black hole.
    This version uses OBSERVER'S UNITS.
    MM = mass  (in msun)
    jj = dimensionless angular momentum parameter
    ell,mm = mode ... (2,2) is the default
    nn = tone ... 0 is the default
    returns:  f_lmn, tau_lmn
    f_lmn is observed frequency in Hz
    tau_lmn is the damping time in seconds
    """
    MM = utils.m_sol_to_geo(MM)
    omega, tau = freq_damping(MM, jj, ell, mm, nn)
    f = omega / (2*np.pi)
    f = utils.freq_geo_to_Hz(f)
    tau = utils.time_geo_to_s(tau)
    return f, tau


def freq_damping(MM, jj, ell=2, mm=2, nn=0):
    """
    Calculate the quasinormal mode freq and damping time for a black hole.
    MM = mass  (in geometric units)
    jj = dimensionless angular momentum parameter
    ell,mm = mode ... (2,2) is the default
    nn = tone ... 0 is the default
    returns:  omega_lmn, tau_lmn
    omega_lmn is ANGULAR frequency in GEOMETRIC units
    tau_lmn is the damping time in geometric units
    """
    data_file = os.path.join(pkg_resources.resource_filename(__name__, 'data'),
                             'fitcoeffsWEB.dat')
    data = np.loadtxt(data_file)

    ell_data = data[:, 0].astype(int)
    mm_data = data[:, 1].astype(int)
    nn_data = data[:, 2].astype(int)

    cond = (ell_data == ell) & (mm_data == mm) & (nn_data == nn)

    f1 = data[cond, 3][0]
    f2 = data[cond, 4][0]
    f3 = data[cond, 5][0]

    q1 = data[cond, 6][0]
    q2 = data[cond, 7][0]
    q3 = data[cond, 8][0]

    # dimensionless frequency
    F_lmn = f1 + f2 * (1. - jj)**f3

    # angular frequency
    omega_lmn = F_lmn / MM

    # quality factor
    Q_lmn = q1 + q2 * (1. - jj)**q3

    # damping time
    tau_lmn = 2. * Q_lmn / omega_lmn

    return omega_lmn, tau_lmn


def amplitude(m1, m2, ell=2, mm=2):
    """
    calculate the amplitude of the models relative to the 22 mode.
    see equations (5) -- (8) from Gossan et al. (2012) (note there is a
    difference between the arXiv paper and the real paper)
    m1, m2 are component masses of the progenitor
    """

    from gwmemory.utils import m12_to_symratio

    # symmetric mass ratio
    nu = m12_to_symratio(m1, m2)

    # all modes are normalised to the A22 mode
    A22 = 0.864 * nu

    if mm == 1:
        if ell == 2:
            return 0.52 * (1 - 4*nu) ** 0.71 * A22
    elif ell == 2 & mm == 2:
        return A22
    elif ell == 3 & mm == 3:
        return 0.44 * (1 - 4*nu) ** 0.45 * A22
    elif ell == 4 & mm == 4:
        return (5.4 * (nu - 0.22)**2 + 0.04) * A22
    else:
        print('Unknown mode ({}, {}) specified'.format(ell, mm))


def final_mass_spin(m1, m2):
    """
    given initial total mass and mass ratio, calculate the final mass, M_f
    and dimensionless spin parameter, jj,
    using fits in Buonanno et al. (2007) -- see the caption of their Table I.
    """

    m1 = float(m1)
    m2 = float(m2)

    eta = m1 * m2 / (m1 + m2)**2
    MM = m1 + m2

    # expansion coefficients for the mass
    M_f1 = 1.
    M_f2 = (np.sqrt(8./9.) - 1.) * eta
    M_f3 = -0.498 * eta**2

    # final mass
    M_f = MM * (M_f1 + M_f2 + M_f3)

    # expansion coefficients for spin parameter
    a_f1 = np.sqrt(12.) * eta
    a_f2 = -2.9 * eta**2

    # final spin parameter -- dimensions of mass
    a_f = M_f * (a_f1 + a_f2)

    # dimensionless spin
    jj = a_f / M_f

    return M_f, jj


def hp_hx(time, m1, m2, ell, mm, iota, phi, phase):
    """
    output hplus and hcross times distance:
    both calculated using eqns. (1) and (2) from Gossan et al.
    calculates for a single ll and mm, can then sum
    iota: angle between spin axis and line-of-sight to observer
    phi: azimuth angle of BH with respect to observer
    """

    Ylm_plus = sYlm(-2, ll=ell, mm=mm, theta=iota, phi=0) + (-1.)**ell * sYlm(-2, ll=ell, mm=-mm, theta=iota, phi=0)
    Ylm_cross = sYlm(-2, ll=ell, mm=mm, theta=iota, phi=0) - (-1.)**ell * sYlm(-2, ll=ell, mm=-mm, theta=iota, phi=0)

    Alm = amplitude(m1, m2, ell=ell, mm=mm)

    M_f, jj = final_mass_spin(m1, m2)

    omega_lm, tau_lm = freq_damping(M_f, jj, ell=ell, mm=mm, nn=0)

    h_plus = M_f * Alm * np.exp(-time / tau_lm) * Ylm_plus * np.cos(omega_lm * time - mm * phi + phase)

    h_cross = M_f * Alm * np.exp(-time / tau_lm) * Ylm_cross * np.sin(omega_lm * time - mm * phi + phase)

    return np.real(h_plus), np.real(h_cross)


def hp_hx_template(time, omega, tau, ell, mm, iota, phi, phase):
    """
    output hplus and hcross template:
    both calculated using eqns. (1) and (2) from Gossan et al.
    but with the normalisations ignored -- i.e., A_{ell m}=1, M_f = 1
    iota: angle between spin axis and line-of-sight to observer
    phi: azimuth angle of BH with respect to observer
    """

    Ylm_plus = sYlm(-2, ll=ell, mm=mm, theta=iota, phi=0) + (-1.)**ell * sYlm(-2, ll=ell, mm=-mm, theta=iota, phi=0)
    Ylm_cross = sYlm(-2, ll=ell, mm=mm, theta=iota, phi=0) - (-1.)**ell * sYlm(-2, ll=ell, mm=-mm, theta=iota, phi=0)

    h_plus = np.exp(-time / tau) * Ylm_plus * np.cos(omega * time - mm * phi + phase)

    h_cross = np.exp(-time / tau) * Ylm_cross * np.sin(omega * time - mm * phi + phase)

    return np.real(h_plus), np.real(h_cross)



from __future__ import absolute_import

from gwmemory import waveforms, utils
import inspect


def time_domain_memory(model=None, h_lm=None, times=None, q=None, MTot=None, S1=None, S2=None, distance=None,
                       inc=None, pol=None, **kwargs):
    """
    Calculate the time domain memory waveform according to __reference__.
    
    Example usage:
    
    Using NR surrogate waveform __reference__ for an edge-on non-spinning, equal-mass, binary
    at a distance of 400 Mpc.
    
    h_mem, times = time_domain_memory(model='NRSur7dq2', q=1, MTot=60, distance=400, inc=np.pi/2, pol=0)
    

    Using an EOBNR waveform __reference__ for an edge-on non-spinning, equal-mass, binary
    at a distance of 400 Mpc.
    
    h_mem, times = time_domain_memory(model='SEOBNRv4', q=1, MTot=60, distance=400, inc=np.pi/2, pol=0)
    

    Using the minimal waveform model __reference__ for an edge-on non-spinning, equal-mass, binary
    at a distance of 400 Mpc.
    
    h_mem, times = time_domain_memory(model='MWM', q=1, MTot=60, distance=400, inc=np.pi/2, pol=0)
    
    
    Using a pre-computed spherical harmonic decomposed waveform for an edge-on non-spinning, equal-mass,
    binary at a distance of 400 Mpc.
    
    h_mem, times = time_domain_memory(h_lm=h_lm, times=times, distance=400, inc=np.pi/2, pol=0)
    
    
    Parameters
    ----------
    model: str
        Name of the model, this is used to identify waveform approximant, e.g., NRSur7dq2, IMRPhenomD, MWM, etc.
    h_lm: dict
        Spin weighted spherical harmonic decomposed time series.
        If this is specified these polarisations will be used.
    times: array
        time series corresponding to the h_lm.
    q: float
        Mass ratio of the binary being considered.
    MTot: float
        Total mass of the binary being considered in solar units.
    S1: array
        Dimensionless spin vector of the more massive black hole.
    S2: array
        Dimensionless spin vector of the less massive black hole.
    distance: float
        Distance to the binary in Mpc.
    inc: float
        Inclination of the binary to the line of sight.
        If not provided, spherical harmonic modes will be returned.
    pol: float
        Orbital polarisation.
        If not provided, spherical harmonic modes will be returned.
    kwargs: dict
        Additional model-specific keyword arguments.
        
    Returns
    -------
    h_mem, dict
        Memory time series, either in spherical harmonic modes or plus/cross polarisations.
    times, array
        Time series corresponding to the memory waveform.
    """
    if h_lm is not None and times is not None:
        wave = waveforms.MemoryGenerator(name=model, h_lm=h_lm, times=times)
    elif 'NRSur' in model:
        model_kwargs = {key: kwargs[key] for key in inspect.getargspec(waveforms.Surrogate.__init__)[0] if key in kwargs}
        wave = waveforms.Surrogate(q=q, name=model, MTot=MTot, S1=S1, S2=S2, distance=distance, times=times, **model_kwargs)
    elif 'EOBNR' in model or 'Phenom' in model:
        model_kwargs = {key: kwargs[key] for key in inspect.getargspec(waveforms.Approximant.__init__)[0] if key in kwargs}
        wave = waveforms.Approximant(q=q, name=model, MTot=MTot, S1=S1, S2=S2, distance=distance, times=times, **model_kwargs)
    elif model == 'MWM':
        model_kwargs = {key: kwargs[key] for key in inspect.getargspec(waveforms.MWM.__init__)[0] if key in kwargs}
        wave = waveforms.MWM(q=q, name=model, MTot=MTot, distance=distance, times=times, **model_kwargs)
    else:
        print('Model {} unknown'.format(model))
        return None

    function_kwargs = {key: kwargs[key] for key in inspect.getargspec(wave.time_domain_memory)[0] if key in kwargs}
    h_mem, times = wave.time_domain_memory(inc=inc, phase=pol, **function_kwargs)
    
    return h_mem, times


def frequency_domain_memory(model=None, q=None, MTot=None, S1=None, S2=None, distance=None, inc=None, pol=None, **kwargs):
    """
    Calculate the frequency domain memory waveform according to __reference__.
    
    Parameters
    ----------
    model: str
        Name of the model, this is used to identify waveform approximant, e.g., NRSur7dq2, IMRPhenomD, MWM, etc.
    h_lm: dict
        Spin weighted spherical harmonic decomposed time series.
        If this is specified these polarisations will be used.
    times: array
        time series corresponding to the h_lm.
    q: float
        Mass ratio of the binary being considered.
    MTot: float
        Total mass of the binary being considered in solar units.
    S1: array
        Dimensionless spin vector of the more massive black hole.
    S2: array
        Dimensionless spin vector of the less massive black hole.
    distance: float
        Distance to the binary in Mpc.
    inc: float
        Inclination of the binary to the line of sight.
        If not provided, spherical harmonic modes will be returned.
    pol: float
        Orbital polarisation.
        If not provided, spherical harmonic modes will be returned.
    kwargs: dict
        Additional model-specific keyword arguments.
        
    Returns
    -------
    h_mem, dict
        Memory frequency series, either in spherical harmonic modes or plus/cross polarisations.
    frequencies, array
        Frequency series corresponding to the memory waveform.
    """
    time_domain_strain, times = time_domain_memory(model=model, q=q, MTot=MTot, S1=S1, S2=S2, distance=distance,
                                                   inc=inc, pol=pol, **kwargs)
    sampling_frequency = 1 / (times[1] - times[0])
    
    frequency_domain_strain = dict()
    for key in time_domain_strain:
        frequency_domain_strain[key], frequencies = utils.nfft(time_domain_strain[key], sampling_frequency)
        
    return frequency_domain_strain, frequencies


# ----------------------------------------------------------
#
# This module computes spin-weighted spherical harmonics.
#
# Released under the MIT License.
# (C) Christian Reisswig 2009-2011
# Modified Colm Talbot 2018
#
# ----------------------------------------------------------

import numpy as np


def fac(n):
    result = 1
    for i in range(2, n+1):
        result *= i
    return result


# coefficient function
def Cslm(s, l, m):
    return np.sqrt(l*l * (4.0*l*l - 1.0) / ((l*l - m*m) * (l*l - s*s)))


# recursion function
def s_lambda_lm(s, l, m, x):
    Pm = pow(-0.5, m)
    if m != s:
        Pm = Pm * pow(1.0+x, (m-s)*1.0/2)
    if m != -s:
        Pm = Pm * pow(1.0-x, (m+s)*1.0/2)
    Pm = Pm * np.sqrt(fac(2*m + 1) * 1.0 / (4.0*np.pi * fac(m+s) * fac(m-s)))
    if l == m:
        return Pm
    Pm1 = (x + s*1.0/(m+1)) * Cslm(s, m+1, m) * Pm
    if l == m+1:
        return Pm1
    else:
        Pn = 0
        for n in range(m+2, l+1):
            Pn = (x + s*m * 1.0 / (n * (n-1.0))) * Cslm(s, n, m) * Pm1 - Cslm(s, n, m) * 1.0 / Cslm(s, n-1, m) * Pm
            Pm = Pm1
            Pm1 = Pn
        return Pn


def sYlm(ss, ll, mm, theta, phi):
    """Calculate spin-weighted harmonic"""
    Pm = 1.0
    l = ll
    m = mm
    s = ss
    if l < 0:
        return 0
    if abs(m) > l or l < abs(s):
        return 0
    if abs(mm) < abs(ss):
        s = mm
        m = ss
    if (m + s) % 2:
        Pm = -Pm
    if m < 0:
        s = -s
        m = -m
    if (m + s) % 2:
        Pm = -Pm
    result = Pm * s_lambda_lm(s, l, m, np.cos(theta))
    return result * np.cos(mm*phi) + result * np.sin(mm*phi) * 1j


def lmax_modes(lmax):
    """Compute all (l, m) pairs with 2<=l<=lmax"""
    return [(l, m) for l in range(2, lmax+1) for m in range(-l, l+1)]



#!/usr/bin/env python

from setuptools import setup

setup(name='gwmemoritz',
      version='0.1.1',
      packages=['gwmemory'],
      package_dir={'gwmemory': 'gwmemory'},
      package_data={'gwmemory': ['data/gamma_coefficients*.dat', 'data/*WEB.dat']},
      install_requires=['numpy', 'scipy', 'pandas', 'glob2', 'deepdish', 'lalsuite', 'NRSur7dq2']
      )



from __future__ import print_function, division, absolute_import

from gwmemory import angles, qnms, utils, harmonics
import numpy as np
import lal
import lalsimulation as lalsim
import NRSur7dq2
from scipy.interpolate import interp1d


class MemoryGenerator(object):

    def __init__(self, name, h_lm=None, times=None, distance=None):
        self.name = name
        self.h_lm = h_lm
        self.times = times
        self.distance = distance

    @property
    def modes(self):
        return self.h_lm.keys()

    @property
    def delta_t(self):
        return self.times[1] - self.times[0]

    def time_domain_memory(self, inc=None, phase=None, gamma_lmlm=None):
        """
        Calculate the spherical harmonic decomposition of the nonlinear memory from a dictionary of spherical mode time
        series

        Parameters
        ----------
        inc: float, optional
            Inclination of the source, if None, the spherical harmonic modes will be returned.
        phase: float, optional
            Reference phase of the source, if None, the spherical harmonic modes will be returned.
            For CBCs this is the phase at coalesence.
        gamma_lmlm: dict
            Dictionary of arrays defining the angular dependence of the different memory modes, default=None
            if None the function will attempt to load them

        Return
        ------
        h_mem_lm: dict
            Time series of the spherical harmonic decomposed memory waveform.
        times: array_like
            Time series on which memory is evaluated.
        """
        if self.h_lm is None:
            _ = self.time_domain_oscillatory()
        lms = self.modes

        dhlm_dt = dict()
        for lm in lms:
            dhlm_dt[lm] = np.gradient(self.h_lm[lm], self.delta_t)

        dhlm_dt_sq = dict()
        for lm in lms:
            for lmp in lms:
                index = (lm, lmp)
                dhlm_dt_sq[index] = dhlm_dt[lm] * np.conjugate(dhlm_dt[lmp])

        if gamma_lmlm is None:
            gamma_lmlm = angles.load_gamma()

        # constant terms in SI units
        const = 1 / 4 / np.pi
        if self.distance is not None:
            const *= self.distance * utils.Mpc / utils.cc

        dh_mem_dt_lm = dict()
        for ii, ell in enumerate(gamma_lmlm['0'].l):
            if ell > 4:
                continue
            for delta_m in gamma_lmlm.keys():
                if abs(int(delta_m)) > ell:
                    continue
                dh_mem_dt_lm[(ell, int(delta_m))] = np.sum(
                    [dhlm_dt_sq[((l1, m1), (l2, m2))] * gamma_lmlm[delta_m]['{}{}{}{}'.format(l1, m1, l2, m2)][ii]
                     for (l1, m1), (l2, m2) in dhlm_dt_sq.keys() if m1 - m2 == int(delta_m)], axis=0)

        h_mem_lm = {lm: const * np.cumsum(dh_mem_dt_lm[lm]) * self.delta_t for lm in dh_mem_dt_lm}

        if inc is None or phase is None:
            return h_mem_lm, self.times
        else:
            return combine_modes(h_mem_lm, inc, phase), self.times

    def time_domain_oscillatory(self, **kwargs):
        pass

    def set_time_array(self, times):
        """
        Change the time array on which the waveform is evaluated.
        
        Parameters
        ----------
        times: array_like
            New time array for waveform to be evaluated on.
        """
        for mode in self.modes:
            interpolated_mode = interp1d(self.times, self.h_lm)
            self.h_lm[mode] = interpolated_mode[times]
        self.times = times


class Surrogate(MemoryGenerator):
    """
    Memory generator for a numerical relativity surrogate.

    Attributes
    ----------
    name: str
        Name of file to extract waveform from.
    modes: dict
        Spherical harmonic modes which we have knowledge of, default is ell<=4.
    h_lm: dict
        Spherical harmonic decomposed time-domain strain.
    times: array
        Array on which waveform is evaluated.
    q: float
        Binary mass ratio
    MTot: float, optional
        Total binary mass in solar units.
    distance: float, optional
        Distance to the binary in Mpc.
    S1: array
        Spin vector of more massive black hole.
    S2: array
        Spin vector of less massive black hole.
    """

    def __init__(self, q, name='', MTot=None, S1=None, S2=None, distance=None, LMax=4, modes=None, times=None):
        """
        Initialise Surrogate MemoryGenerator
        
        Parameters
        ----------
        name: str
            File name to load.
        LMax: int
            Maximum ell value for oscillatory time series.
        modes: dict, optional
            Modes to load in, default is all ell<=4.
        q: float
            Binary mass ratio
        MTot: float, optional
            Total binary mass in solar units.
        distance: float, optional
            Distance to the binary in Mpc.
        S1: array_like
            Spin vector of more massive black hole.
        S2: array_like
            Spin vector of less massive black hole.
        times: array_like
            Time array to evaluate the waveforms on, default is np.linspace(-900, 100, 10001).
        """
        MemoryGenerator.__init__(self, name=name, distance=distance)
        self.sur = NRSur7dq2.NRSurrogate7dq2()

        self.q = q
        self.MTot = MTot
        self.S1 = S1
        self.S2 = S2
        self.LMax = LMax
        self.times = times
        self.h_lm, self.times = self.time_domain_oscillatory(modes=modes, times=self.geometric_times)

    def time_domain_oscillatory(self, times=None, modes=None, inc=None, phase=None):
        """
        Get the mode decomposition of the surrogate waveform.

        Calculates a BBH waveform using the surrogate models of Field et al. (2014), Blackman et al. (2017)
        http://journals.aps.org/prx/references/10.1103/PhysRevX.4.031006, https://arxiv.org/abs/1705.07089
        See https://data.black-holes.org/surrogates/index.html for more information.

        Parameters
        ----------
        times: np.array, optional
            Time array (In geometric units) on which to evaluate the waveform.
        modes: list, optional
            List of modes to try to generate.
        inc: float, optional
            Inclination of the source, if None, the spherical harmonic modes will be returned.
        phase: float, optional
            Phase at coalescence of the source, if None, the spherical harmonic modes will be returned.

        Returns
        -------
        h_lm: dict
            Spin-weighted spherical harmonic decomposed waveform.
        times: np.array
            Times on which waveform is evaluated.
        """
        if self.h_lm is None:
            if times is None:
                times = self.default_geometric_times
            times = times * self.geo_to_t
            h_lm = self.sur(self.q, self.S1, self.S2, MTot=self.MTot, distance=self.distance, t=times, LMax=self.LMax)

            available_modes = set(h_lm.keys())

            if modes is None:
                modes = available_modes

            if not set(modes).issubset(available_modes):
                print('Requested {} unavailable modes'.format(' '.join(set(modes).difference(available_modes))))
                modes = list(set(modes).union(available_modes))
                print('Using modes {}'.format(' '.join(modes)))

            h_lm = {(ell, m): h_lm[ell, m] for ell, m in modes}

        else:
            h_lm = self.h_lm
            times = self.times

        if inc is None or phase is None:
            return h_lm, times
        else:
            return combine_modes(h_lm, inc, phase), times

    @property
    def times(self):
        return self.__times

    @times.setter
    def times(self, times):
        self.__times = times
        if times is None:
            return
        if max(self.geometric_times) > max(self.default_geometric_times):
            print("Warning: Time array exceeds the maximum allowed by NRSurrogate.\n Geometric time array max {} > "
                  "NRSurrogate max {}".format(max(self.geometric_times), max(self.default_geometric_times)))
        if min(self.geometric_times) > min(self.default_geometric_times):
            print("Warning: Time array exceeds the minimum allowed by NRSurrogate.\n Geometric time array min {} > "
                  "NRSurrogate min {}".format(min(self.geometric_times), min(self.default_geometric_times)))

    @property
    def default_geometric_times(self):
        return np.linspace(-900, 100, 10001)

    @property
    def q(self):
        return self.__q

    @q.setter
    def q(self, q):
        if q < 1:
            q = 1 / q
        if q > 2:
            print('WARNING: Surrogate waveform not tested for q>2.')
        self.__q = q

    @property
    def S1(self):
        return self.__s1

    @S1.setter
    def S1(self, S1):
        if S1 is None:
            self.__s1 = np.array([0., 0., 0.])
        else:
            self.__s1 = np.array(S1)

    @property
    def S2(self):
        return self.__s2

    @S2.setter
    def S2(self, S2):
        if S2 is None:
            self.__s2 = np.array([0., 0., 0.])
        else:
            self.__s2 = np.array(S2)

    @property
    def h_to_geo(self):
        if self.MTot is None:
            return 1
        else:
            return self.distance * utils.Mpc / self.MTot / utils.solar_mass / utils.GG * utils.cc ** 2

    @property
    def t_to_geo(self):
        if self.MTot is None:
            return None
        else:
            return 1 / self.MTot / utils.solar_mass / utils.GG * utils.cc ** 3

    @property
    def geo_to_t(self):
        return 1 / self.t_to_geo

    @property
    def geometric_times(self):
        if self.times is not None:
            return self.times * self.t_to_geo
        else:
            return None


class SXSNumericalRelativity(MemoryGenerator):
    """
    Memory generator for a numerical relativity waveform.

    Attributes
    ----------
    name: str
        Name of file to extract waveform from.
    modes: dict
        Spherical harmonic modes which we have knowledge of, default is ell<=4.
    h_lm: dict
        Spherical harmonic decomposed time-domain strain.
    times: array
        Array on which waveform is evaluated.
    MTot: float, optional
        Total binary mass in solar units.
    distance: float, optional
        Distance to the binary in Mpc.
    """

    def __init__(self, name, modes=None, extraction='OutermostExtraction.dir', MTot=None, distance=None, times=None):
        """
        Initialise SXSNumericalRelativity MemoryGenerator
        
        Parameters
        ----------
        name: str
            File name to load.
        modes: dict, optional
            Modes to load in, default is all ell<=4.
        extraction: str
            Extraction method, this specifies the outer object to use in the h5 file.
        MTot: float
            Lab-frame total mass of the binary in solar masses.
        distance: float
            Luminosity distance to the binary in Mpc.
        times: array_like
            Time array to evaluate the waveforms on, default is time array in h5 file.
        """
        self.name = name
        self.h_lm, self.times = utils.load_sxs_waveform(name, modes=modes, extraction=extraction)

        self.MTot = MTot
        self.distance = distance

        if MTot is None or distance is None:
            self.h_to_geo = 1
            self.t_to_geo = 1
        else:
            self.h_to_geo = self.distance * utils.Mpc / self.MTot / utils.solar_mass / utils.GG * utils.cc ** 2
            self.t_to_geo = 1 / self.MTot / utils.solar_mass / utils.GG * utils.cc ** 3

            for mode in self.h_lm:
                self.h_lm /= self.h_to_geo
            self.times / self.t_to_geo
            # Rezero time array to the merger time
            self.times -= self.times[np.argmax(abs(self.h_lm[(2, 2)]))]

        if times is not None:
            self.set_time_array(times)

        MemoryGenerator.__init__(self, name=name, h_lm=self.h_lm, times=times, distance=self.distance)

    def time_domain_oscillatory(self, times=None, modes=None, inc=None, phase=None):
        """
        Get the mode decomposition of the numerical relativity waveform.

        Parameters
        ----------
        inc: float, optional
            Inclination of the source, if None, the spherical harmonic modes will be returned.
        phase: float, optional
            Phase at coalescence of the source, if None, the spherical harmonic modes will be returned.

        Returns
        -------
        h_lm: dict
            Spin-weighted spherical harmonic decomposed waveform.
        times: np.array
            Times on which waveform is evaluated.
        """
        if inc is None or phase is None:
            return self.h_lm, times
        else:
            return combine_modes(self.h_lm, inc, phase), times


class Approximant(MemoryGenerator):

    def __init__(self, name, q, MTot=60, S1=np.array([0, 0, 0]), S2=np.array([0, 0, 0]), distance=400, times=None,
                 inc=0, phase=0):
        """
        Initialise Surrogate MemoryGenerator
        
        Parameters
        ----------
        name: str
            File name to load.
        q: float
            Binary mass ratio
        MTot: float, optional
            Total binary mass in solar units.
        distance: float, optional
            Distance to the binary in Mpc.
        S1: array_like
            Spin vector of more massive black hole.
        S2: array_like
            Spin vector of less massive black hole.
        times: array_like
            Time array to evaluate the waveforms on, default is time array from lalsimulation.
            FIXME
        """
        self.q = q
        self.MTot = MTot
        self.__S1 = S1
        self.__S2 = S2
        self._check_prececssion()

        MemoryGenerator.__init__(self, name=name, times=times, distance=distance)


    @property
    def available_modes(self):
        return [(2, 2), (2, -2)]

    @property
    def m1(self):
        return self.MTot / (1 + self.q)

    @property
    def m2(self):
        return self.m1 * self.q

    @property
    def m1_SI(self):
        return self.m1 * utils.solar_mass

    @property
    def m2_SI(self):
        return self.m2 * utils.solar_mass

    @property
    def distance_SI(self):
        return self.distance * utils.Mpc

    @property
    def q(self):
        return self.__q

    @q.setter
    def q(self, q):
        if q > 1:
            q = 1 / q
        self.__q = q

    @property
    def h_to_geo(self):
        return self.distance_SI / (self.m1_SI + self.m2_SI) / utils.GG * utils.cc ** 2

    @property
    def t_to_geo(self):
        return 1 / (self.m1_SI + self.m2_SI) / utils.GG * utils.cc ** 3

    @property
    def delta_t(self):
        return self.times[1] - self.times[0]

    @property
    def S1(self):
        return self.__S1

    @S1.setter
    def S1(self, S1):
        if S1 is None:
            self.__S1 = np.array([0., 0., 0.])
        else:
            self.__S1 = np.array(S1)
        self._check_prececssion()

    @property
    def S2(self):
        return self.__S2

    @S2.setter
    def S2(self, S2):
        if S2 is None:
            self.__S2 = np.array([0., 0., 0.])
        else:
            self.__S2 = np.array(S2)
        self._check_prececssion()

    @property
    def approximant(self):
        return lalsim.GetApproximantFromString(self.name)

    def _check_prececssion(self):
        if abs(self.__S1[0]) > 0 or abs(self.__S1[1]) > 0 or abs(self.__S2[0]) > 0 or abs(self.__S2[1]) > 0:
            print('WARNING: Approximant decomposition works only for non-precessing waveforms.')
            print('Setting spins to be aligned')
            self.__S1[0], self.__S1[1] = 0., 0.
            self.__S2[0], self.__S2[1] = 0., 0.
            print('New spins are: S1 = {}, S2 = {}'.format(self.__S1, self.__S2))
        else:
            self.__S1 = list(self.__S1)
            self.__S2 = list(self.__S2)

    def time_domain_oscillatory(self, modes=None, inc=None, phase=None):
        """
        Get the mode decomposition of the waveform approximant.

        Since the waveforms we consider only contain content about the ell=|m|=2 modes.
        We can therefore evaluate the waveform for a face-on system, where only the (2, 2) mode
        is non-zero.

        Parameters
        ----------
        modes: list, optional
            List of modes to try to generate.
        inc: float, optional
            Inclination of the source, if None, the spherical harmonic modes will be returned.
        phase: float, optional
            Phase at coalescence of the source, if None, the spherical harmonic modes will be returned.

        Returns
        -------
        h_lm: dict
            Spin-weighted spherical harmonic decomposed waveform.
        times: np.array
            Times on which waveform is evaluated.
        """
        if self.h_lm is None:
            if modes is None:
                modes = self.available_modes
            else:
                modes = modes

            if not set(modes).issubset(self.available_modes):
                print('Requested {} unavailable modes'.format(' '.join(set(modes).difference(self.available_modes))))
                modes = list(set(modes).union(self.available_modes))
                print('Using modes {}'.format(' '.join(modes)))

            fmin = 20
            fRef = 20
            theta = 0.0
            phi = 0.0
            longAscNodes = 0.0
            eccentricity = 0.0
            meanPerAno = 0.0
            WFdict = lal.CreateDict()

            hplus, hcross = lalsim.SimInspiralChooseTDWaveform(
                self.m1_SI, self.m2_SI, self.S1[0], self.S1[1], self.S1[2], self.S2[0], self.S2[1], self.S2[2],
                self.distance_SI, theta, phi, longAscNodes, eccentricity, meanPerAno, self.delta_t, fmin, fRef,
                WFdict, self.approximant)

            h = hplus.data.data - 1j * hcross.data.data
            h_22 = h / harmonics.sYlm(-2, 2, 2, theta, phi)

            self.h_lm = {(2, 2): h_22, (2, -2): np.conjugate(h_22)}
            self.zero_pad_h_lm()

        if inc is None or phase is None:
            return self.h_lm
        else:
            return combine_modes(h_lm=self.h_lm, inc=inc, phase=phase)

    def zero_pad_h_lm(self):
        required_zeros = len(self.times) - len(self.h_lm.values()[0])
        if required_zeros == 0:
            return
        elif required_zeros > 0:
            for mode in self.h_lm:
                result = np.zeros(self.times.shape, dtype=np.complex128)
                result[:self.h_lm[mode].shape[0]] = self.h_lm[mode]
                self.h_lm[mode] = result
        else:
            raise ValueError("The defined time array is too short!")


class MWM(MemoryGenerator):

    def __init__(self, q, MTot=60, distance=400, name='MWM', times=None):
        MemoryGenerator.__init__(self, name=name, h_lm=dict(), times=times, distance=distance)
        self.name = name
        if q > 1:
            q = 1 / q
        self.q = q
        self.MTot = MTot
        self.m1 = self.MTot / (1 + self.q)
        self.m2 = self.m1 * self.q

        self.h_to_geo = self.distance * utils.Mpc / (self.m1 + self.m2) / utils.solar_mass \
                        / utils.GG * utils.cc ** 2
        self.t_to_geo = 1 / (self.m1 + self.m2) / utils.solar_mass / utils.GG * utils.cc ** 3

        if times is None:
            times = np.linspace(-900, 100, 10001) / self.t_to_geo
        self.times = times

    def time_domain_memory(self, inc, phase, times=None, rm=3):
        """
        Calculates the plus and cross polarisations for the
        minimal waveform model memory waveform:
        eqns (5) and (9) from Favata (2009. ApJL 159)

        TODO: Implement spherical harmonic decomposition?

        Parameters
        ----------
        inc: float
            Binary inclination angle
        phase: float
            Binary phase at coalscence
        times: array, optional
            Time array on which the memory is calculated
        rm: float, optional
            Radius at which the matching occurs in solar masses

        Returns
        -------
        h_mem: dict
            Plus and cross polarisations of the memory waveform
        times: np.array
            Time array on which the memory is calculated

        Paul Lasky
        """
        if times is None:
            times = self.times

        time_geo = utils.time_s_to_geo(times)  # units: metres

        m1_geo = utils.m_sol_to_geo(self.m1)  # units: metres
        m2_geo = utils.m_sol_to_geo(self.m2)  # units: metres

        dist_geo = utils.dist_Mpc_to_geo(self.distance)  # units: metres

        # total mass
        MM = m1_geo + m2_geo

        # symmetric mass ratio
        eta = utils.m12_to_symratio(m1_geo, m2_geo)

        # this is the orbital separation at the matching radius -- see Favata (2009) before eqn (8).
        # the default value for this is given as rm = 3 MM.
        rm *= MM

        # calculate dimensionless mass and spin of the final black hole
        # from the Buonanno et al. (2007) fits
        Mf_geo, jj = qnms.final_mass_spin(m1_geo, m2_geo)

        # calculate the QNM frequencies and damping times
        # from the fits in Table VIII of Berti et al. (2006)
        omega220, tau220 = qnms.freq_damping(Mf_geo, jj, ell=2, mm=2, nn=0)
        omega221, tau221 = qnms.freq_damping(Mf_geo, jj, ell=2, mm=2, nn=1)
        omega222, tau222 = qnms.freq_damping(Mf_geo, jj, ell=2, mm=2, nn=2)

        sigma220 = 1j * omega220 + 1 / tau220
        sigma221 = 1j * omega221 + 1 / tau221
        sigma222 = 1j * omega222 + 1 / tau222

        # set the matching time to be at t = 0
        tm = 0
        TT = time_geo - tm

        # some quantity defined after equation (7) of Favata
        trr = 5 * MM * rm ** 4 / (256 * eta * MM ** 4)

        # calculate the A_{ell m n} matching coefficients.  Note that I've solved
        # a matrix equation that solved for the three coefficients fron three equations
        xi = 2 * np.sqrt(2 * np.pi / 5) * eta * MM * rm ** 2
        chi = -2 * 1j * np.sqrt(MM / rm ** 3)

        A220 = xi * (sigma221 * sigma222 * chi ** 2 + sigma221 * chi ** 3 + sigma222 * chi ** 3 + chi ** 4) \
               / ((sigma220 - sigma221) * (sigma220 - sigma222))
        A221 = xi * (sigma220 * sigma222 * chi ** 2 + sigma220 * chi ** 3 + sigma222 * chi ** 3 + chi ** 4) \
               / ((sigma221 - sigma220) * (sigma221 - sigma222))
        A222 = xi * (sigma220 * sigma221 * chi ** 2 + sigma220 * chi ** 3 + sigma221 * chi ** 3 + chi ** 4) \
               / ((sigma221 - sigma222) * (sigma220 - sigma222))

        # Calculate the coefficients in the summed term of equation (9) from Favata (2009)
        # this is a double sum, with each variable going from n = 0 to 2; therefore 9 terms
        coeffSum00 = sigma220 * np.conj(sigma220) * A220 * np.conj(A220) / (sigma220 + np.conj(sigma220))
        coeffSum01 = sigma220 * np.conj(sigma221) * A220 * np.conj(A221) / (sigma220 + np.conj(sigma221))
        coeffSum02 = sigma220 * np.conj(sigma222) * A220 * np.conj(A222) / (sigma220 + np.conj(sigma222))

        coeffSum10 = sigma221 * np.conj(sigma220) * A221 * np.conj(A220) / (sigma221 + np.conj(sigma220))
        coeffSum11 = sigma221 * np.conj(sigma221) * A221 * np.conj(A221) / (sigma221 + np.conj(sigma221))
        coeffSum12 = sigma221 * np.conj(sigma222) * A221 * np.conj(A222) / (sigma221 + np.conj(sigma222))

        coeffSum20 = sigma222 * np.conj(sigma220) * A222 * np.conj(A220) / (sigma222 + np.conj(sigma220))
        coeffSum21 = sigma222 * np.conj(sigma221) * A222 * np.conj(A221) / (sigma222 + np.conj(sigma221))
        coeffSum22 = sigma222 * np.conj(sigma222) * A222 * np.conj(A222) / (sigma222 + np.conj(sigma222))

        # radial separation
        rr = rm * (1 - TT / trr) ** (1 / 4)

        # set up strain
        h_MWM = np.zeros(len(TT))

        # calculate strain for TT < 0.
        h_MWM[TT <= 0] = 8 * np.pi * MM / rr[TT <= 0]

        # calculate strain for TT > 0.
        term00 = coeffSum00 * (1 - np.exp(-TT[TT > 0] * (sigma220 + np.conj(sigma220))))
        term01 = coeffSum01 * (1 - np.exp(-TT[TT > 0] * (sigma220 + np.conj(sigma221))))
        term02 = coeffSum02 * (1 - np.exp(-TT[TT > 0] * (sigma220 + np.conj(sigma222))))

        term10 = coeffSum10 * (1 - np.exp(-TT[TT > 0] * (sigma221 + np.conj(sigma220))))
        term11 = coeffSum11 * (1 - np.exp(-TT[TT > 0] * (sigma221 + np.conj(sigma221))))
        term12 = coeffSum12 * (1 - np.exp(-TT[TT > 0] * (sigma221 + np.conj(sigma222))))

        term20 = coeffSum20 * (1 - np.exp(-TT[TT > 0] * (sigma222 + np.conj(sigma220))))
        term21 = coeffSum21 * (1 - np.exp(-TT[TT > 0] * (sigma222 + np.conj(sigma221))))
        term22 = coeffSum22 * (1 - np.exp(-TT[TT > 0] * (sigma222 + np.conj(sigma222))))

        sum_terms = np.real(term00 + term01 + term02 +
                            term10 + term11 + term12 +
                            term20 + term21 + term22)

        h_MWM[TT > 0] = 8 * np.pi * MM / rm + sum_terms / (eta * MM)

        # calculate the plus polarisation of GWs: eqn. (5) from Favata (2009)
        sT = np.sin(inc)
        cT = np.cos(inc)

        h_plus_coeff = 0.77 * eta * MM / (384 * np.pi) * sT ** 2 * (17 + cT ** 2) / dist_geo
        h_mem = dict(plus=h_plus_coeff * h_MWM, cross=np.zeros_like(h_MWM))

        return h_mem, times


def combine_modes(h_lm, inc, phase):
    """Calculate the plus and cross polarisations of the waveform from the spherical harmonic decomposition."""
    total = sum([h_lm[(l, m)] * harmonics.sYlm(-2, l, m, inc, phase) for l, m in h_lm])
    h_plus_cross = dict(plus=total.real, cross=-total.imag)
    return h_plus_cross
